"""
Copyright (C) 2012 Haak Saxberg

This file is part of Flowhub, a command-line tool to enable various
Git-based workflows that interacts with GitHub.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

from ConfigParser import DuplicateSectionError
from collections import OrderedDict
import re


class ImproperlyConfigured(Exception):
    pass


class Configurator(object):
    """Supports configuration files with subsections, given git.config.write
    instances"""
    def __init__(self, config_object):
        self._confer = config_object
        self._read_only = config_object.read_only

        self._sections = OrderedDict()

        for section_name in self._confer.sections():
            match = re.match('(?P<section>.+) "(?P<subsection>.+)"$', section_name)
            if match:

                supersection_n = match.groupdict()['section']
                section_n = match.groupdict()['subsection']

                supersection = self._sections.setdefault(supersection_n, Section(supersection_n, self, read_only=self._read_only))

                try:
                    section = supersection.add_section(section_n)
                except AttributeError as e:
                    print e
                    section = getattr(supersection, section_n)

            else:
                section = self._sections.setdefault(section_name, Section(section_name, self, read_only=self._read_only))

            try:
                for value_name, value in self._confer._sections[section_name].iteritems():
                    section._values[value_name] = value
            except AttributeError as e:
                print e

    def add_section(self, section_name):
        self._confer.add_section(section_name)

        match = re.match('(?P<section>.+) "(?P<subsection>.+)"$', section_name)
        if match:

            supersection_n = match.groupdict()['section']
            section_n = match.groupdict()['subsection']

            supersection = self._sections.setdefault(supersection_n, Section(supersection_n, self, read_only=self._read_only))

            section = supersection._subsections.setdefault(section_n, Section(section_name, self, read_only=self._read_only))
        else:
            section = self._sections.setdefault(section_name, Section(section_name, self, read_only=self._read_only))

        return section

    def write(self):
        self._confer.write()

    def __getattr__(self, attr):
        if attr in self._sections:
            return self._sections[attr]

        return object.__getattribute__(self, attr)


class Section(object):
    """Dotted-syntax access to nested settings"""
    def __init__(self, name, configurator, read_only=False, parent=None):
        """
            name: a string
            configurator: a Configurator instance
            read_only: boolean; is this a read_only section
            parent: None or a Section instance
        """

        self._name = name
        self._configurator = configurator
        self._subsections = OrderedDict()
        self._values = OrderedDict()
        self._read_only = read_only
        self._parent = parent

    def add_section(self, section_name):
        if section_name in self._subsections:
            raise DuplicateSectionError(section_name)

        section = Section(section_name, self._configurator, read_only=self._read_only, parent=self)
        self._subsections[section_name] = section

        return section

    def set_value(self, value_name, value):
        if self._read_only:
            raise AttributeError("This is a read-only instance.")

        self._values[value_name] = value
        self._configurator._confer.set(self._name, value_name, value)
        self._configurator._confer.write()

    def __getattr__(self, attr):
        if attr in super(Section, self).__getattribute__('_subsections'):
            return super(Section, self).__getattribute__('_subsections')[attr]

        elif attr in super(Section, self).__getattribute__('_values'):
            return super(Section, self).__getattribute__('_values')[attr]

        return super(Section, self).__getattribute__(attr)

    def __setattr__(self, attr, value):
        try:
            if attr in self._subsections:
                raise RuntimeError("Can't overwrite subsections this way.")

            elif not attr.startswith('_'):
                self.set_value(attr, value)

            else:
                super(Section, self).__setattr__(attr, value)
        except AttributeError:
            super(Section, self).__setattr__(attr, value)

    def __repr__(self):
        return '<Section: {}>'.format(self._name)



"""
Copyright (C) 2012 Haak Saxberg

This file is part of Flowhub, a command-line tool to enable various
Git-based workflows that interacts with GitHub.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
from collections import namedtuple

TagInfo = namedtuple("TagInfo", ["label", "message"])


class Manager(object):

    def __init__(
        self,
        debug,
        prefix,
        origin,
        canon,
        master,
        develop,
        release,
        hotfix,
        repo,
        gh,
        offline
    ):
        self._prefix = prefix
        self.DEBUG = debug
        self.origin = origin
        self.canon = canon
        self.master = master
        self.develop = develop
        self.release = release
        self.hotfix = hotfix
        self.repo = repo
        self.gh = gh
        self.offline = offline





#!/usr/bin/env python
"""
Copyright (C) 2012 Haak Saxberg

This file is part of Flowhub, a command-line tool to enable various
Git-based workflows that interacts with GitHub.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
import argparse
import os
import subprocess
import tempfile

from engine import Engine
from managers import TagInfo


__version__ = "0.5.1a"


def future_proof_print(x):
    print(x)


def do_hook(args, engine, hook_name, *hook_args):
    if args.no_verify:
        return True

    try:
        hook_args = tuple(str(a) for a in hook_args)
        subprocess.check_call((os.path.join(engine._repo.git_dir, 'hooks', hook_name),) + hook_args)
        return True
    except OSError as e:
        if args.verbosity > 2:
            print "No such hook: {}".format(hook_name)
            print "({})".format(e)
        return True
    except subprocess.CalledProcessError:
        return False


def create_tag_info(args, default_label=""):
    label = raw_input("Tag Label [{}]: ".format(default_label)) or default_label

    # Open the $EDITOR, if you can...
    descr_f = tempfile.NamedTemporaryFile(delete=False)
    descr_f.file.write(
        "\n\n# Write the tag description above"
    )
    if args.verbosity > 3:
        print "Temp file: ", descr_f.name
    # regardless, close the tempfile.
    descr_f.close()

    try:
        editor_result = subprocess.check_call(
            "$EDITOR {}".format(descr_f.name),
            shell=True
        )
    except OSError:
        if args.verbosity > 2:
            print "Hmm...are you on Windows?"
        editor_result = 126

    if args.verbosity > 3:
        print "result of $EDITOR: ", editor_result

    if editor_result == 0:
        # Re-open the file to get new contents...
        fnew = open(descr_f.name, 'r')
        # and remove the first line
        body = fnew.readlines()
        if body[-1].startswith('# Write the tag description'):
            body = body[:-1]

        body = "".join(body)

        fnew.close()
    else:
        body = raw_input(
            "Tag message:\n"
        )

    return TagInfo(label.strip(), body.strip())


def handle_init_call(args, engine, input_func=raw_input, output_func=future_proof_print):
    if args.verbosity > 2:
        output_func("handling init")

    # doesn't do anything but setup.
    # Setup was already done, during engine construction.
    if args.verbosity > 2:
        output_func("Begin repo setup")

    name = input_func(
        "Name of the GitHub repository for this flowhub: "
    )

    origin = input_func("Name of your github remote [origin]: ") or 'origin'
    # if not engine._remote_exists(origin):
    #     print "Whoops! That remote doesn't exist."
    #     remote_url = input_func("Remote url: ")
    #     engine._repo.create_remote(
    #         origin,
    #         remote_url,
    #     )
    canon = input_func('Name of the organization remote [canon]: ') or 'canon'
    # if not engine._remote_exists(canon):
    #     print "Whoops! That remote doesn't exist."
    #     remote_url = input_func("Remote url: ")
    #     self._repo.create_remote(
    #         canon,
    #         remote_url,
    #     )

    master = input_func("Name of the stable branch [master]: ") or 'master'
    if not engine._branch_exists(master):
        output_func("\tCreating branch {}".format(master))
        engine._repo.create_head(master)

    develop = input_func("Name of the development branch [develop]: ") or 'develop'
    if not engine._branch_exists(master):
        output_func("\tCreating branch {}".format(develop))
        engine._repo.create_head(master)

    feature = input_func("Prefix for feature branches [feature/]: ") or 'feature/'
    release = input_func("Prefix for release branches [release/]: ") or "release/"
    hotfix = input_func("Prefix for hotfix branches [hotfix/]: ") or "hotfix/"

    engine.setup_repository_structure(
        name,
        origin,
        canon,
        master,
        develop,
        feature,
        release,
        hotfix,
    )


def handle_feature_call(args, engine):
    if args.verbosity > 2:
        print "handling feature"
    if args.action == 'start':
        if args.issue_number:
            args.name = "{}-{}".format(args.issue_number, args.name)
        engine.create_feature(
            name=args.name,
            with_tracking=args.track,
        )
        do_hook(args, engine, "post-feature-start")

    elif args.action == 'work':
        if not args.issue:
            engine.work_feature(name=args.identifier)
        else:
            engine.work_feature(issue=args.identifier)

    elif args.action == 'publish':
        if not do_hook(args, engine, "pre-feature-publish"):
            return False
        try:
            engine.publish_feature(name=args.name)
        except AssertionError:
            # This is janky as shit, but running twice fixes it.
            # see #14
            engine.publish_feature(name=args.name)

    elif args.action == 'abandon':
        engine.abandon_feature(
            name=args.name,
        )

    elif args.action == 'accepted':
        try:
            engine.accept_feature(
                name=args.name,
                delete_feature_branch=(not args.no_delete),
            )
        except AssertionError:
            # see #14
            engine.accept_feature(
                name=args.name,
                delete_feature_branch=(not args.no_delete),
            )
    elif args.action == 'list':
        engine.list_features()
    else:
        raise RuntimeError("Unimplemented command for features: {}".format(args.action))


def handle_hotfix_call(args, engine):
    if args.verbosity > 2:
        print "handling hotfix"

    if args.action == 'start':
        engine.start_hotfix(
            name=args.name,
            issues=args.issue_numbers,
        )
        do_hook(args, engine, "post-hotfix-start", args.name)
    elif args.action == 'publish':
        if not engine.hotfix:
            return False

        if not do_hook(args, engine, "pre-hotfix-publish"):
            return False

        default_tag = engine.hotfix.name.replace(
            engine.hotfix_manager._prefix, ""
        )
        results = engine.publish_hotfix(
            name=args.name,
            tag_info=create_tag_info(args, default_tag),
        )

        do_hook(args, engine, "post-hotfix-publish", results)

    elif args.action == 'contribute':
        engine.contribute_hotfix()
    else:
        raise RuntimeError("Unimplemented command for hotfixes: {}".format(args.action))


def handle_release_call(args, engine):
    if args.verbosity > 2:
        print "handling release"

    if args.action == 'start':
        engine.start_release(
            name=args.name,
        )
        do_hook(args, engine, "post-release-start", args.name)

    elif args.action == 'publish':
        if not engine.release:
            return False

        if not do_hook(args, engine, "pre-release-publish"):
            return False

        default_tag = engine.release.name.replace(
            engine.release_manager._prefix, ""
        )
        results = engine.publish_release(
            name=args.name,
            tag_info=create_tag_info(args, default_tag),
            with_delete=(not args.no_cleanup),
        )
        do_hook(args, engine, "post-release-publish", results)

    elif args.action == 'contribute':
        engine.contribute_release()
    else:
        raise RuntimeError("Unimplemented command for releases: {}".format(args.action))


def handle_cleanup_call(args, engine):
    if args.verbosity > 2:
        print "handling cleanup"

    # Get the targets for cleanup
    targets = ''
    if args.t or args.all:
        if args.verbosity > 2:
            print "adding 't' to targets"
        targets += 't'
    if args.u or args.all:
        if args.verbosity > 2:
            print "adding 'u' to targets"
        targets += 'u'
    if args.r or args.all:
        if args.verbosity > 2:
            print "adding 'r' to targets"
        targets += 'r'

    if targets:
        engine.cleanup_branches(targets=targets)
    else:
        print "No targets specified for cleanup."


def handle_issue_call(args, engine):
    if args.verbosity > 2:
        print "handling issue call"

    if args.action == 'start':
        engine.open_issue(
            title=args.title,
            labels=args.labels.split(',') if args.labels else None,
            create_branch=args.create_branch,
        )


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', '--verbosity', action="store", type=int, default=0)
    parser.add_argument('--offline', action='store_true', default=False,
        help='do not talk to GitHub',)
    parser.add_argument('--no-verify', action='store_true', default=False,
        help='do not call any hooks',)
    parser.add_argument('--version', action='version',
        version=('flowhub v{}'.format(__version__)))

    subparsers = parser.add_subparsers(dest="subparser")
    subparsers.add_parser('init',
        help="set up a repository to use flowhub",)
    feature = subparsers.add_parser('feature',
        help="do feature-related things",)
    hotfix = subparsers.add_parser('hotfix',
        help="do hotfix-related things",)
    release = subparsers.add_parser('release',
        help="do release-related things",)
    cleanup = subparsers.add_parser('cleanup',
        help="do repository-cleanup related things",)
    issue = subparsers.add_parser('issue',
        help="do issue-related things",)

    #
    # Features
    #
    feature_subs = feature.add_subparsers(dest='action')

    fstart = feature_subs.add_parser('start',
        help="start a new feature branch")
    fstart.add_argument('name', help="name of the feature")
    fstart.add_argument('--track', default=False, action='store_true',
        help="set up a tracking branch on your github immediately.")
    fstart.add_argument('-i', '--issue-number', type=int,
        action='store', default=None,
        help="prepend an issue number to the feature name")

    fwork = feature_subs.add_parser('work',
        help="switch to a different feature (by name)")
    fwork.add_argument('identifier', help="name of feature to switch to")
    fwork.add_argument('--issue', '-i',
        action='store_true', default=False,
        help='switch to a branch by issue number instead of by name')

    fpublish = feature_subs.add_parser('publish',
        help="send the current feature branch to origin and create a pull-request")
    fpublish.add_argument('name', nargs='?',
        default=None,
        help='name of feature to publish. If not given, uses current feature')

    fabandon = feature_subs.add_parser('abandon',
        help="remove a feature branch completely"
    )
    fabandon.add_argument('name', nargs='?',
        default=None,
        help="name of the feature to abandon. If not given, uses current feature")
    faccepted = feature_subs.add_parser('accepted',
        help="declare that a feature was accepted into the trunk")
    faccepted.add_argument('name', nargs='?',
        default=None,
        help="name of the accepted feature. If not given, assumes current feature")
    faccepted.add_argument('--no-delete', action='store_true', default=False,
        help="don't delete the accepted feature branch")
    feature_subs.add_parser('list',
        help='list the feature names on this repository')

    #
    # Hotfixes
    #
    hotfix_subs = hotfix.add_subparsers(dest='action')

    hstart = hotfix_subs.add_parser('start',
        help="start a new hotfix branch")
    hstart.add_argument('name',
        help="name (and tag) for the hotfix")
    hstart.add_argument('--issue-numbers', '-i', type=int,
        default=None, nargs='+',
        help="specifies the issues this hotfix addresses")
    hpublish = hotfix_subs.add_parser('publish',
        help="publish the hotfix to production and trunk")
    hpublish.add_argument('name', nargs='?',
        help="name of hotfix to publish. If not given, uses current branch.")
    hotfix_subs.add_parser('contribute',
        help='send this branch as a pull request to the current hotfix')
    #
    # Releases
    #
    release_subs = release.add_subparsers(dest='action')

    rstart = release_subs.add_parser('start',
        help="start a new release branch")
    rstart.add_argument('name', help="name (and tag) of the release branch.")

    release_subs.add_parser('stage',
        help="send a release branch to a staging environment")

    rpublish = release_subs.add_parser('publish',
        help="publish a release branch to production and trunk")
    rpublish.add_argument('name', nargs='?',
        help="name of release to publish. if not specified, current branch is assumed.")
    rpublish.add_argument('--no-cleanup', action='store_true',
        default=False,
        help="do not delete the release branch after a successful publish",
    )
    release_subs.add_parser('contribute')

    # rabandon = release_subs.add_parser('abandon',
    #     help='abort a release branch')
    # rabandon.add_argument('name', nargs='?',
    #     help='name of release to abandon. if not specified, current branch is assumed.')

    #
    # Cleanup
    #
    cleanup.add_argument('-u', action='store_true',
        help='do cleanup features.', default=False)
    cleanup.add_argument('-r', action='store_true',
        help='do cleanup releases.', default=False)
    cleanup.add_argument('-t', action='store_true',
        help='do cleanup hotfixes.', default=False)
    cleanup.add_argument('-a', '--all', action='store_true', default=False,
        help='Shorthand for using the flag -urt')

    #
    # Issues
    #
    issue_subs = issue.add_subparsers(dest='action')
    istart = issue_subs.add_parser('start',
        help="Open a new issue on github")
    istart.add_argument('title', nargs='?', default=None, action='store',
        help="Title of the created issue")
    istart.add_argument('--labels', '-l', default=None, action='store',
        help='Comma-separated list of labels to apply to this bug.\nLabels that don\'t exist won\'t be applied.')
    istart.add_argument('--create-branch', '-b', default=False, action='store_true',
        help="Create a feature branch for this issue.")

    args = parser.parse_args()
    if args.verbosity > 2:
        print "Args: ", args

    # Force initialization to run offline.
    if args.subparser == 'init':
        e = Engine(debug=args.verbosity, offline=True)
        handle_init_call(args, e)
        return

    else:
        e = Engine(debug=args.verbosity, offline=args.offline)

    if args.subparser == 'feature':
        handle_feature_call(args, e)

    elif args.subparser == 'hotfix':
        handle_hotfix_call(args, e)

    elif args.subparser == 'release':
        handle_release_call(args, e)

    elif args.subparser == 'cleanup':
        handle_cleanup_call(args, e)

    elif args.subparser == 'issue':
        handle_issue_call(args, e)

    else:
        raise RuntimeError("Unrecognized command: {}".format(args.subparser))

if __name__ == "__main__":
    run()



"""
Copyright (C) 2012 Haak Saxberg

This file is part of Flowhub, a command-line tool to enable various
Git-based workflows that interacts with GitHub.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import git
import mock
import os
from unittest import TestCase

from flowhub.configurator import Configurator, DuplicateSectionError, Section
from tests import TEST_REPO
from tests import id_generator


class SectionTestCase(TestCase):
    def do_set_up(self, **kwargs):
        self.mock_configurator = mock.MagicMock()
        self.section = Section(
            kwargs.get('name', id_generator()),
            self.mock_configurator,
            kwargs.get('read_only', False),
            kwargs.get('parent', None),
        )


class SectionWriteEnabledTestCase(SectionTestCase):
    def setUp(self):
        self.do_set_up()

    def test_initialize(self):
        pass

    def test_add_section(self):
        section_name = id_generator()
        new_section = self.section.add_section(section_name)

        self.assertEqual(
            new_section._parent, self.section
        )

        self.assertIn(
            new_section._name,
            self.section._subsections,
        )
        self.assertEqual(
            self.section._subsections.get(section_name),
            new_section,
        )

    def test_add_dup_section(self):
        section_name = id_generator()
        self.section.add_section(section_name)

        with self.assertRaises(DuplicateSectionError):
            self.section.add_section(section_name)

    def test_set_value(self):
        value_name = id_generator()
        value = id_generator()

        self.assertFalse(hasattr(self.section, value_name))

        self.section.set_value(value_name, value)

        self.assertTrue(hasattr(self.section, value_name))
        self.assertEqual(getattr(self.section, value_name), value)

        self.mock_configurator.assert_has_calls([
            mock.call._confer.set(self.section._name, value_name, value),
            mock.call._confer.write()
        ])

    def test_set_value_dot_syntax(self):
        value_name = id_generator()
        value = id_generator()
        self.section.set_value(value_name, value)

        new_value = "NEW" + str(value)
        setattr(self.section, value_name, new_value)

        self.assertEqual(getattr(self.section, value_name), new_value)

    def test_invalid_subsection_overwrite(self):
        value = id_generator()
        section_name = id_generator()
        self.section.add_section(section_name)

        with self.assertRaises(RuntimeError):
            setattr(self.section, section_name, value)


class SectionWriteDisabledTestCase(SectionTestCase):
    def setUp(self):
        self.do_set_up(read_only=True)

    def test_initialization(self):
        pass

    def test_set_value(self):
        value_name = id_generator()
        value = id_generator()

        with self.assertRaises(AttributeError):
            self.section.set_value(value_name, value)


class ConfiguratorTestCase(TestCase):
    def setUp(self):
        print "Creating new test repo..."
        self.repo = git.Repo.init(TEST_REPO)
        # make an initial commit
        self.repo.index.commit("Initial commit")
        os.chdir(TEST_REPO)

        self._setup_configurator()

    def _setup_configurator(self, **kwargs):
        self.sections = [
            "core",
            "flowhub \"auth\"",
            "flowhub \"structure\"",
        ]
        self.mock_config = mock.MagicMock(
            read_only=kwargs.get('read_only', False),
            sections=lambda: self.sections,
        )
        self.configurator = Configurator(self.mock_config)

    def _test_basic_sections(self, configurator):

        self.assertTrue(hasattr(configurator, 'core'))
        self.assertTrue(hasattr(configurator, 'flowhub'))
        self.assertTrue(hasattr(configurator.flowhub, 'auth'))
        self.assertTrue(hasattr(configurator.flowhub, 'structure'))


class ConfiguratorInitializationTestCase(ConfiguratorTestCase):

    def test_reader_init(self):
        self._test_basic_sections(self.configurator)

    def test_writer_init(self):
        self._test_basic_sections(self.configurator)


class ConfiguratorAddSectionTestCase(ConfiguratorTestCase):
    def test_add_section(self):
        section_name = id_generator()

        self.configurator.add_section(section_name)

        self.assertTrue(hasattr(self.configurator, section_name))
        self.mock_config.assert_has_calls([
            mock.call.add_section(section_name),
        ])

    def test_add_section_with_subsection(self):
        section_name = id_generator()
        subsection_name = id_generator()

        section_whole_name = "{} \"{}\"".format(section_name, subsection_name)
        self.configurator.add_section(section_whole_name)

        self.assertTrue(hasattr(self.configurator, section_name))
        self.assertTrue(hasattr(getattr(self.configurator, section_name), subsection_name))

    def test_add_duplicate_section(self):
        section_name = id_generator()

        self.configurator.add_section(section_name)
        self.configurator.add_section(section_name)



"""
Copyright (C) 2012 Haak Saxberg

This file is part of Flowhub, a command-line tool to enable various
Git-based workflows that interacts with GitHub.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import getpass
import re
import subprocess
import tempfile
import warnings

import git
from github import Github, GithubException

from configurator import Configurator, ImproperlyConfigured
from decorators import with_summary
from managers import TagInfo
from managers.feature import FeatureManager
from managers.hotfix import HotfixManager
from managers.pull_request import PullRequestManager
from managers.release import ReleaseManager


class NoSuchObject(Exception):
    pass

class NoSuchBranch(NoSuchObject):
    pass
class NoSuchRemote(NoSuchObject):
    pass


def online_only(method):
    def wrapper(self, *args, **kwargs):
        if self.offline:
            print "not available offline"
            return False

        return method(self, *args, **kwargs)

    return wrapper


class Engine(object):
    def __init__(self, debug=0, INIT=False, offline=False):
        self.DEBUG = debug
        if self.DEBUG > 2:
            print "initing engine"

        # assume flowhub is called from within a git repository
        self._repo = git.Repo(".")
        self._cr = Configurator(self._repo.config_reader())

        self._gh = None

        self.offline = offline
        if not self.offline:
            if self.DEBUG > 0:
                print "Authorizing engine..."
            if not self.do_auth():
                print "Authorization failed! Exiting."
                return

            try:
                self._gh_repo = self._gh.get_user().get_repo(
                    self._cr.flowhub.structure.name
                )
            except GithubException:
                raise ImproperlyConfigured(
                    "No repo with given name: {}".format(
                        self._cr.flowhub.structure.name,
                    )
                )

            if self._gh.rate_limiting[0] < 100:
                warnings.warn(
                    "You are close to exceeding your GitHub access rate; {} left out of {}".format(
                        *self._gh.rate_limiting
                    )
                )
        else:
            if self.DEBUG > 0:
                print "Skipping auth - GitHub accesses will fail."

        if not INIT:
            self.feature_manager = FeatureManager(
                debug=self.DEBUG,
                prefix=self._cr.flowhub.prefix.feature,
                origin=self.origin,
                canon=self.canon,
                master=self.master,
                develop=self.develop,
                release=self.release,
                hotfix=self.hotfix,
                repo=self._repo,
                gh=self._gh,
                offline=self.offline,
            )

            self.release_manager = ReleaseManager(
                debug=self.DEBUG,
                prefix=self._cr.flowhub.prefix.release,
                origin=self.origin,
                canon=self.canon,
                master=self.master,
                develop=self.develop,
                release=self.release,
                hotfix=self.hotfix,
                repo=self._repo,
                gh=self._gh,
                offline=self.offline,
            )

            self.hotfix_manager = HotfixManager(
                debug=self.DEBUG,
                prefix=self._cr.flowhub.prefix.hotfix,
                origin=self.origin,
                canon=self.canon,
                master=self.master,
                develop=self.develop,
                release=self.release,
                hotfix=self.hotfix,
                repo=self._repo,
                gh=self._gh,
                offline=self.offline,
            )

            self.pull_manager = PullRequestManager(
                debug=self.DEBUG,
                prefix=self._cr.flowhub.structure.name,
                origin=self.origin,
                canon=self.canon,
                master=self.master,
                develop=self.develop,
                release=self.release,
                hotfix=self.hotfix,
                repo=self._repo,
                gh=self._gh,
                offline=self.offline,
            )

    def do_auth(self):
        """Generates the authorization to do things with github."""
        try:
            token = self._cr.flowhub.auth.token
            self._gh = Github(token)
            if self.DEBUG > 0:
                print "GitHub Engine authorized by token in settings."
        except AttributeError:
            print (
                "Flowhub needs permission to access your GitHub repositories.\n"
                "Entering your credentials now will grant Flowhub the access it "
                "requires."
            )
            if not self._create_token():
                return False
            # Refresh the readers
            self._cr = Configurator(self._repo.config_reader())

        return True

    def _create_token(self):
        # Don't store the users' information.
        for i in range(3):
            self._gh = Github(raw_input("Username: "), getpass.getpass())

            try:
                auth = self._gh.get_user().create_authorization(
                    ['user', 'repo', 'gist'],
                    'Flowhub Client',
                )
                break
            except GithubException:
                print "Invalid username/password combination."
                if i == 2:
                    return False

        token = auth.token
        if self.DEBUG > 2:
            print "Token generated: ", token
        # set the token globally, rather than on the repo level.
        authing = subprocess.check_output(
            'git config --global --add flowhub.auth.token {}'.format(token),
            shell=True,
        ).strip()
        if self.DEBUG > 2:
            print "result of config set:", authing

        return True

    def setup_repository_structure(
        self,
        name,
        origin,
        canon,
        master,
        develop,
        feature,
        release,
        hotfix,
    ):
        cw = self._repo.config_writer()
        if self.DEBUG > 2:
            print "Begin repo setup"
        if not cw.has_section('flowhub "structure"'):
            cw.add_section('flowhub "structure"')

        cw.set('flowhub "structure"', 'name', name)

        cw.set('flowhub "structure"', 'origin', origin)
        cw.set('flowhub "structure"', 'canon', canon)

        cw.set('flowhub "structure"', 'master', master)

        if not self._branch_exists(master):
            print "\tCreating branch {}".format(master)
            self._repo.create_head(master)

        cw.set('flowhub "structure"', 'develop', develop)
        if not self._branch_exists(develop):
            print "\tCreating branch {}".format(develop)
            self._repo.create_head(develop)

        if not cw.has_section('flowhub "prefix"'):
            cw.add_section('flowhub "prefix"')

        cw.set('flowhub "prefix"', 'feature', feature)
        cw.set('flowhub "prefix"', 'release', release)
        cw.set('flowhub "prefix"', 'hotfix', hotfix)
        cw.write()

        # Refresh the read-only reader.
        self._cr = Configurator(self._repo.config_reader())

    def _branch_exists(self, branch_name):
        if self.DEBUG > 2:
            print "Checking for existence of branch {}".format(branch_name)
        return getattr(self._repo.heads, branch_name, None) is not None

    def _remote_exists(self, repo_name):
        if self.DEBUG > 2:
            print "Checking for existence of remote {}".format(repo_name)
        return getattr(self._repo.remotes, repo_name, None) is not None

    def __get_branch_by_name(self, name):
        try:
            return getattr(self._repo.heads, name)
        except AttributeError:
            raise NoSuchBranch(name)

    @property
    def develop(self):
        develop_name = self._cr.flowhub.structure.develop
        if self.DEBUG > 3:
            print "finding develop branch {}".format(develop_name)
        return self.__get_branch_by_name(develop_name)

    @property
    def master(self):
        master_name = self._cr.flowhub.structure.master
        if self.DEBUG > 3:
            print "finding master branch {}".format(master_name)
        return self.__get_branch_by_name(master_name)

    def __get_remote_by_name(self, name):
        try:
            return getattr(self._repo.remotes, name)
        except AttributeError:
            raise NoSuchRemote(name)

    @property
    def origin(self):
        origin_name = self._cr.flowhub.structure.origin
        if self.DEBUG > 3:
            print "finding origin repo {}".format(origin_name)
        return self.__get_remote_by_name(origin_name)

    @property
    def canon(self):
        canon_name = self._cr.flowhub.structure.canon
        if self.DEBUG > 3:
            print "finding canon repo {}".format(canon_name)
        return self.__get_remote_by_name(canon_name)

    @property
    def gh_canon(self):
        # if this isn't a fork, we have slightly different sha's.
        if self.canon == self.origin:
            gh_parent = self._gh_repo
        else:
            gh_parent = self._gh_repo.parent

        return gh_parent

    @property
    def release(self):
        # official version releases are named release/#.#.#
        releases = [x for x in self._repo.branches if x.name.startswith(
            self._cr.flowhub.prefix.release,
        )]

        if releases:
            return releases[0]
        else:
            return None

    @property
    def hotfix(self):
        # official version hotfixes are named release/#.#.#
        hotfixes = [x for x in self._repo.branches if x.name.startswith(
            self._cr.flowhub.prefix.hotfix,
        )]

        if hotfixes:
            return hotfixes[0]
        else:
            return None

    def _create_pull_request(self, base, head, summary):

        # try to glean issue numbers from branch
        pr_from_issue = self.pull_manager.create_from_branch_name(base, head, summary)
        if pr_from_issue:
            return pr_from_issue

        is_issue = raw_input("is this feature answering an issue? [y/N] ").lower() == 'y'
        if not is_issue:
            issue = self._open_issue(summary=summary, return_issue=True)

            if self.DEBUG > 1:
                print (issue.title, issue.body, base, head)

        else:
            good_number = False
            while not good_number:
                try:
                    issue_number = int(raw_input("issue number: "))
                except ValueError:
                    print "not a valid number"
                    continue

                issue = self.pull_manager.get_issue(issue_number)
                if issue is None:
                    print "no such issue"
                    continue

                good_number = True

        pr = self.pull_manager.create_pull(
            issue=issue,
            base=base,
            head=head,
            summary=summary,
        )

        return pr

    def _create_feature(self, name=None, with_tracking=True, summary=None):
        if name is None:
            print "please provide a feature name."
            return False

        if summary is None:
            summary = []

        branch = self.feature_manager.start(
            name,
            with_tracking,
            summary,
        )

        branch.checkout()

        summary += [
            "Checked out branch {}".format(branch.name),
        ]

        return True
    create_feature = with_summary(_create_feature)

    def work_feature(self, name=None):
        """Simply checks out the feature branch for the named feature."""
        if name is None:
            print "please provide a feature name."
            return False

        branches = self.feature_manager.fuzzy_get(name)

        if len(branches) == 1:
            branches[0].checkout()
            print "switched to branch '{}'".format(branches[0].name)

        elif len(branches) > 1:
            print "multiple branches found:"
            for branch in branches:
                print "\t{}".format(branch)

        else:
            print "No feature starts with {}".format(name)

        return True

    def _accept_feature(self, name=None, delete_feature_branch=True, summary=None):
        if summary is None:
            summary = []

        return_branch = self._repo.head.reference
        if name is None:
            # If no name specified, try to use the currently checked-out branch,
            # but only if it's a feature branch.
            name = self._repo.head.reference.name
            if self._cr.flowhub.prefix.feature not in name:
                print (
                    "Please provide a feature name, or switch to "
                    "the feature branch you want to mark as accepted."
                )
                return False

            name = name.replace(self._cr.flowhub.prefix.feature, '')
            return_branch = self.develop

        self.feature_manager.accept(
            name,
            summary=summary,
            with_delete=delete_feature_branch,
        )

        return_branch.checkout()
        summary += [
            "Checked out branch {}".format(return_branch.name),
        ]

        return True
    accept_feature = with_summary(_accept_feature)

    def _abandon_feature(self, name=None, summary=None):
        if summary is None:
            summary = []
        return_branch = self._repo.head.reference
        if name is None:
            # If no name specified, try to use the currently checked-out branch,
            # but only if it's a feature branch.
            name = self._repo.head.reference.name
            if self._cr.flowhub.prefix.feature not in name:
                print (
                    "Please provide a feature name, or switch to "
                    "the feature branch you want to abandon."
                )
                return False

            name = name.replace(self._cr.flowhub.prefix.feature, '')
            return_branch = self.develop

        if self.DEBUG > 0:
            print "Abandoning feature branch..."

        # checkout develop
        # branch -D feature_prefix+name
        # push --delete origin feature_prefix+name

        return_branch.checkout()
        summary += [
            "Checked out branch {}".format(
                return_branch.name,
            ),
        ]

        self.feature_manager.abandon(
            name,
            summary=summary,
        )
        return True
    abandon_feature = with_summary(_abandon_feature)

    @online_only
    def _publish_feature(self, name=None, summary=None):
        if summary is None:
            summary = []
        if name is None:
            # If no name specified, try to use the currently checked-out branch,
            # but only if it's a feature branch.
            name = self._repo.head.reference.name
            if self._cr.flowhub.prefix.feature not in name:
                print (
                    "please provide a feature name, or switch to "
                    "the feature branch you want to publish."
                )
                return False

            name = name.replace(self._cr.flowhub.prefix.feature, '')

        branch = self.feature_manager.publish(name, summary)

        # we don't have access to gh_canon if we're offline
        if not self.offline:
            base = self.develop
            pr = self.pull_manager.add_to_pull(base, branch, summary)

            if not pr:
                pr = self._create_pull_request(base, branch, summary)

        return True
    publish_feature = with_summary(_publish_feature)

    def list_features(self):
        features = [
            b for b in self._repo.branches
            if b.name.startswith(self._cr.flowhub.prefix.feature)
        ]
        if not features:
            print "There are no feature branches."
            return

        for branch in features:
            display = '{}'.format(
                branch.name.replace(
                    self._cr.flowhub.prefix.feature,
                    ''
                ),
            )
            if self._repo.head.reference.name == branch.name:
                display = '* {}'.format(display)
            else:
                display = '  {}'.format(display)

            print display

        return features

    def _start_release(self, name=None, summary=None):
        # checkout develop
        # if already release branch, abort.
        # checkout -b relase_prefix+branch_name

        if summary is None:
            summary = []

        if name is None:
            print "Please provide a release name."
            return False

        if any([
            x for x in self._repo.branches
                if x.name.startswith(self._cr.flowhub.prefix.release)
        ]):
            print "You already have a release in the works - please finish that one."
            return False

        if self.DEBUG > 0:
            print "Creating new release branch..."

        branch = self.release_manager.start(name, summary)

        branch.checkout()

        summary += [
            "Checked out branch {}"
            "\n\nBump the release version now!".format(branch)
        ]

        return True
    start_release = with_summary(_start_release)

    @with_summary
    def stage_release(self, summary=None):
        summary += [
            "Release branch sent off to stage",
        ]
        summary += [
            "Release branch checked out and refreshed on stage."
            "\n\nLOL just kidding, this doesn't do anything."
        ]

    def _publish_release(
        self,
        name=None,
        with_delete=True,
        summary=None,
        tag_info=None,
    ):
        # fetch canon
        # checkout master
        # merge canon master
        # merge --no-ff name
        # tag
        # checkout develop
        # merge canon develop
        # merge --no-ff name
        # push --tags canon
        # delete release branch
        # git push origin --delete name
        return_branch = self._repo.head.reference

        if summary is None:
            summary = []

        if name is None:
            # If no name specified, try to use the currently checked-out branch,
            # but only if it's a feature branch.
            name = self._repo.head.reference.name
            if self._cr.flowhub.prefix.release not in name:
                print (
                    "Please provide a release name, or switch to "
                    "the release branch you want to publish."
                )
                return False

            name = name.replace(self._cr.flowhub.prefix.release, '')
            return_branch = self.develop

        self.release_manager.publish(name, with_delete, tag_info, summary)

        return_branch.checkout()
        summary += [
            "Checked out branch {}".format(return_branch.name),
        ]
        return name
    publish_release = with_summary(_publish_release)

    @online_only
    def _contribute_release(self, summary=None):
        if summary is None:
            summary = []

        if not (self.release and self.release.commit in self._repo.head.reference.object.iter_parents()):
            # Don't allow random branches to be contributed.
            print (
                "You are attempting to contribute a branch that is not a "
                "descendant of a current release.\n"
                "Unfortunately, this isn't allowed."
            )
            return False

        branch = self._repo.head.reference

        self.release_manager.contribute(branch, summary)

        pr = self.pull_manager.add_to_pull(self.release, branch, summary)
        if not pr:
            pr = self._create_pull_request(self.release, branch, summary)

        return True
    contribute_release = with_summary(_contribute_release)

    @with_summary
    def cleanup_branches(self, summary=None, targets=""):
        current_branch = self._repo.head.reference
        hotfix_prefix = self._cr.flowhub.prefix.hotfix
        release_prefix = self._cr.flowhub.prefix.release

        for branch in self._repo.branches:
            if (
                ('u' in targets and branch.name.startswith(self._cr.flowhub.prefix.feature))
                or ('r' in targets and branch.name.startswith(self._cr.flowhub.prefix.release))
                or ('t' in targets and branch.name.startswith(self._cr.flowhub.prefix.hotfix))
            ):
                # Feature branches get removed if they're fully merged in to something else.
                # NOTE: this will delete branch references that have no commits in them.
                if branch == current_branch:
                    print (
                        "Currently checked out branch would be cleaned up; skipping."
                        "If you want this branch to be cleaned up, switch to a different branch"
                        "and re-run this command."
                    )
                    continue

                try:
                    remote_branch = branch.tracking_branch()

                    # If it failed because it's an un-recognizably-merged hotfix
                    # or release contribution, but there's no hotfix/release branch
                    # currently, delete it.
                    if hotfix_prefix in branch.name and not self.hotfix:
                        self._repo.delete_head(branch.name, force=True)
                    elif release_prefix in branch.name and not self.release:
                        self._repo.delete_head(branch.name, force=True)
                    else:
                        self._repo.delete_head(branch.name)
                    summary += [
                        "Deleted local branch {}".format(branch.name)
                    ]

                    if remote_branch:
                        # get rid of the 'origin/' part of the remote name
                        remote_name = '/'.join(remote_branch.name.split('/')[1:])
                        self.origin.push(
                            remote_name,
                            delete=True,
                        )
                        summary[-1] += ' and remote branch {}'.format(remote_branch.name)
                    else:
                        # Sometimes the tracking isn't set properly (at least for empty featuers?)
                        # so, we brute it here.
                        if hasattr(self.origin.refs, branch.name):
                            self.origin.push(
                                branch.name,
                                delete=True,
                            )
                            summary[-1] += '\n\tand remote branch {}/{}'.format(
                                self.origin.name,
                                branch.name,
                            )

                except git.GitCommandError as e:
                    print e
                    continue

    def _start_hotfix(self, name=None, issues=None, summary=None):
        # Checkout master
        # if already hotfix branch, abort.
        # checkout -b hotfix_prefix+branch_name
        if summary is None:
            summary = []

        if name is None:
            print "Please provide a release name."
            return

        if any([
            x for x in self._repo.branches
                if x.name.startswith(self._cr.flowhub.prefix.hotfix)
        ]):
            print (
                "You already have a hotfix in the works - please finish that one."
            )
            return False

        if self.DEBUG > 0:
            print "Creating new hotfix branch..."

        # checkout develop
        # checkout -b hotfix/name

        branch = self.hotfix_manager.start(name, issues, summary)

        # Checkout the branch.
        branch.checkout()
        summary += [
            "Checked out branch {}"
            "\n\nBump the release version now!".format(branch),
        ]
        return True

    start_hotfix = with_summary(_start_hotfix)

    def _publish_hotfix(
        self,
        name=None,
        summary=None,
        with_delete=True,
        tag_info=None,
    ):
        # fetch canon
        # checkout master
        # merge --no-ff hotfix
        # tag
        # checkout develop
        # merge --no-ff hotfix
        # push --tags canon
        # delete hotfix branches
        return_branch = self._repo.head.reference

        if summary is None:
            summary = []
        if name is None:
            # If no name specified, try to use the currently checked-out branch,
            # but only if it's a hotfix branch.
            name = self._repo.head.reference.name
            if self._cr.flowhub.prefix.hotfix not in name:
                print ("please provide a hotfix name, or switch to the hotfix branch you want to publish.")
                return

            name = name.replace(self._cr.flowhub.prefix.hotfix, '')
            return_branch = self.develop

        self.hotfix_manager.publish(name, tag_info, with_delete, summary)

        return_branch.checkout()
        summary += [
            "Checked out branch {}".format(return_branch.name),
        ]

        return name
    publish_hotfix = with_summary(_publish_hotfix)

    @online_only
    def _contribute_hotfix(self, summary=None):
        if not (self.hotfix and self.hotfix.commit in self._repo.head.reference.object.iter_parents()):
            # Don't allow random branches to be contributed.
            print (
                "You are attempting to contribute a branch that is not a "
                "descendant of the current hotfix.\n"
                "Unfortunately, this isn't allowed."
            )
            return False

        branch = self._repo.head.reference

        self.release_manager.contribute(branch, summary)
        pr = self.pull_manager.add_to_pull(self.hotfix, branch, summary)

        if not pr:
            self._create_pull_request(self.hotfix, branch, summary)

        return True
    contribute_release = with_summary(_contribute_release)

    @online_only
    def _open_issue(
        self,
        title=None,
        labels=None,
        create_branch=False,
        summary=None,
        return_issue=False
    ):
        if title is None:
            title = raw_input("Title for this issue: ")
        else:
            print "Title for this issue: ", title

        if labels is None:
            labels = []

        if summary is None:
            summary = []

        # Open the $EDITOR, if you can...
        descr_f = tempfile.NamedTemporaryFile(delete=False)
        descr_f.file.write(
            "\n\n# Write your description above. Remember - you can use GitHub markdown syntax!"
        )
        if self.DEBUG > 3:
            print "Temp file: ", descr_f.name
        # regardless, close the tempfile.
        descr_f.close()

        try:
            editor_result = subprocess.check_call(
                "$EDITOR {}".format(descr_f.name),
                shell=True
            )
        except OSError:
            if self.DEBUG > 2:
                print "Hmm...are you on Windows?"
            editor_result = 126

        if self.DEBUG > 3:
            print "result of $EDITOR: ", editor_result

        if editor_result == 0:
            # Re-open the file to get new contents...
            fnew = open(descr_f.name, 'r')
            # and remove the first line
            body = fnew.readlines()
            if body[-1].startswith('# Write your description'):
                body = body[:-1]

            body = "".join(body)

            fnew.close()
        else:
            body = raw_input(
                "Description (remember, you can use GitHub markdown):\n"
            )

        if self.DEBUG > 3:
            print "Description used:\n", body

        issue = self.pull_manager.open_issue(title, body, labels, summary)

        if create_branch:
            feature_name = "{}-{}".format(
                issue.number,
                title.replace(' ', '-').lower(),
            )
            self.feature_manager.start(feature_name, summary, with_tracking=False)

        if return_issue:
            return issue

    open_issue = with_summary(_open_issue)



"""
Copyright (C) 2012 Haak Saxberg

This file is part of Flowhub, a command-line tool to enable various
Git-based workflows that interacts with GitHub.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""




"""
Copyright (C) 2012 Haak Saxberg

This file is part of Flowhub, a command-line tool to enable various
Git-based workflows that interacts with GitHub.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

from flowhub.managers import Manager


class ReleaseManager(Manager):

    def start(self, name, summary):
        # checkout develop
        # checkout -b release/name

        branch_name = "{}{}".format(
            self._prefix,
            name,
        )
        branch = self.repo.create_head(
            branch_name,
            commit=self.develop,
        )
        summary += [
            "New branch {} created, from branch {}".format(
                branch_name,
                self.develop,
            ),
        ]

        if self.DEBUG > 0:
            print "Adding a tracking branch to your GitHub repo"

        if not self.offline:
            self.canon.push(
                "{0}:{0}".format(branch),
                set_upstream=True,
            )

            summary += [
                "Pushed {} to {}".format(branch, self.canon.name),
            ]

        return branch

    def publish(self, name, with_delete, tag_info, summary):
        if self.offline:
            return False
        release_name = "{}{}".format(
            self._prefix,
            name,
        )

        self.canon.fetch()
        summary += [
            "Latest objects fetched from {}".format(self.canon),
        ]

        # TODO: ensure equality of remote and local master/develop branches
        # TODO: handle merge conflicts.
        # merge into master
        self.master.checkout()
        self.repo.git.merge(
            release_name,
            no_ff=True,
        )
        summary += [
            "Branch {} merged into {}".format(release_name, self.master),
        ]

        # and tag
        if tag_info:
            self.repo.create_tag(
                path=tag_info.label,
                ref=self.master,
                message=tag_info.message,
            )
            summary += [
                "New tag ({}) created at {}'s tip".format(tag_info.label, self.master),
            ]

        # merge into develop
        self.develop.checkout()
        self.repo.git.merge(
            self.master,
            no_ff=True,
        )
        summary += [
            "Branch {} merged into {}".format(self.master, self.develop),
        ]

        # push to canon
        self.canon.push()
        self.canon.push(tags=True)
        summary += [
            "{}, {}, and tags have been pushed to {}".format(
                self.master,
                self.develop,
                self.canon
            ),
        ]

        if with_delete:
            self.repo.delete_head(release_name)
            self.canon.push(
                release_name,
                delete=True,
            )
            summary += [
                "Branch {} removed".format(release_name),
            ]

        return True

    def contribute(self, branch, summary):

        self.repo.git.push(
            self.origin,
            branch,
            set_upstream=True,
        )
        summary += [
            "Branch {} pushed to {}".format(branch, self.origin)
        ]



"""
Copyright (C) 2012 Haak Saxberg

This file is part of Flowhub, a command-line tool to enable various
Git-based workflows that interacts with GitHub.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import itertools
import os
import random
import string
from subprocess import CalledProcessError
import unittest

import mock

from flowhub.core import *
from flowhub.managers import TagInfo
from tests import id_generator


class CoreTestCase(unittest.TestCase):
    def setUp(self):
        self.engine_patch = mock.patch("flowhub.engine.Engine")

        self.args = mock.MagicMock()
        self.engine_mock = self.engine_patch.start()

    def tearDown(self):
        self.engine_patch.stop()


class HookTestCase(CoreTestCase):
    def setUp(self):
        super(HookTestCase, self).setUp()

        self.subprocess_patch = mock.patch('subprocess.check_call')
        self.subprocess_mock = self.subprocess_patch.start()

    def tearDown(self):
        self.subprocess_patch.stop()

        super(HookTestCase, self).tearDown()

    def test_no_verify(self):
        self.args.no_verify = True

        self.assertTrue(do_hook(self.args, self.engine_mock, id_generator()))
        self.assertEqual(self.subprocess_mock.call_count, 0)

    def test_successful_hook(self):
        self.args.no_verify = False
        hook_name = id_generator()
        self.engine_mock._repo.git_dir = id_generator()

        self.assertTrue(do_hook(self.args, self.engine_mock, hook_name))
        self.subprocess_mock.assert_has_calls([
            mock.call(
                ((os.path.join(self.engine_mock._repo.git_dir, 'hooks', hook_name)),),
            ),
        ])

    def test_no_such_hook(self):
        self.args.no_verify = False
        hook_name = id_generator()
        self.subprocess_mock.side_effect = OSError

        self.assertTrue(do_hook(self.args, self.engine_mock, hook_name))

    def test_with_failed_hook(self):
        self.args.no_verify = False
        hook_name = id_generator()
        self.subprocess_mock.side_effect = CalledProcessError(None, None, None)

        self.assertFalse(do_hook(self.args, self.engine_mock, hook_name))


class InitCallTestCase(CoreTestCase):

    def test_correct_setup_defaults(self):
        inputs = ["REPO_NAME", "", "", "", "", "", "", ""]
        input_gen = itertools.islice(inputs, None)

        def input_func(query_str):
            return input_gen.next()

        output_func = lambda x: None

        handle_init_call(self.args, self.engine_mock, input_func=input_func, output_func=output_func)

        self.engine_mock.assert_has_calls([
            mock.call.setup_repository_structure(
                "REPO_NAME",
                "origin",
                "canon",
                "master",
                "develop",
                "feature/",
                "release/",
                "hotfix/",
            ),
        ])

    def test_correct_setup(self):
        inputs = [id_generator() for x in range(8)]
        input_gen = itertools.islice(inputs, None)

        def input_func(query_str):
            return input_gen.next()

        output_func = lambda x: None

        handle_init_call(self.args, self.engine_mock, input_func, output_func)

        self.engine_mock.assert_has_calls([
            mock.call.setup_repository_structure(*inputs)
        ])


class FeatureCallTestCase(CoreTestCase):
    def test_start(self):
        self.args.action = "start"
        self.args.issue_num = None
        self.args.name = id_generator(chars=string.ascii_letters)
        self.args.track = False

        with mock.patch('flowhub.core.do_hook') as patch:
            patch.return_value = True
            handle_feature_call(self.args, self.engine_mock)

            patch.assert_has_calls([
                mock.call(self.args, self.engine_mock, "post-feature-start")
            ])

            self.engine_mock.assert_has_calls([
                mock.call.create_feature(
                    name=self.args.name,
                    with_tracking=self.args.track,
                ),
            ])

    def test_start_with_issue_number(self):
        self.args.action = "start"
        self.args.name = id_generator(chars=string.ascii_letters)
        self.args.issue_num = id_generator(chars=string.digits)
        expected_name = "{}-{}".format(self.args.issue_number, self.args.name)
        self.args.track = False

        handle_feature_call(self.args, self.engine_mock)

        self.engine_mock.assert_has_calls([
            mock.call.create_feature(
                name=expected_name,
                with_tracking=self.args.track,
            ),
        ])

    def test_work(self):
        self.args.action = "work"
        self.args.issue = None
        self.args.identifier = id_generator()
        handle_feature_call(self.args, self.engine_mock)

        self.engine_mock.assert_has_calls([
            mock.call.work_feature(name=self.args.identifier),
        ])

    def test_work_with_issue_num(self):
        self.args.action = "work"
        self.args.issue = True
        self.args.identifier = id_generator()
        handle_feature_call(self.args, self.engine_mock)

        self.engine_mock.assert_has_calls([
            mock.call.work_feature(issue=self.args.identifier),
        ])

    def test_publish(self):
        self.args.action = "publish"
        self.args.name = id_generator()
        with mock.patch('flowhub.core.do_hook') as patch:
            patch.return_value = True

            handle_feature_call(self.args, self.engine_mock)

            patch.assert_has_calls([
                mock.call(self.args, self.engine_mock, "pre-feature-publish")
            ])

            self.engine_mock.assert_has_calls([
                mock.call.publish_feature(name=self.args.name),
            ])

    def test_publish_failed_hook(self):
        self.args.action = "publish"
        with mock.patch('flowhub.core.do_hook') as patch:
            patch.return_value = False
            handle_feature_call(self.args, self.engine_mock)

            self.assertEqual(self.engine_mock.call_count, 0)

    def test_abandon(self):
        self.args.action = "abandon"
        self.args.name = id_generator()
        handle_feature_call(self.args, self.engine_mock)
        self.engine_mock.assert_has_calls([
            mock.call.abandon_feature(name=self.args.name),
        ])

    def test_accepted(self):
        self.args.action = 'accepted'
        self.args.name = id_generator()
        self.args.no_delete = False

        handle_feature_call(self.args, self.engine_mock)

        self.engine_mock.assert_has_calls([
            mock.call.accept_feature(
                name=self.args.name,
                delete_feature_branch=True,
            ),
        ])

    def test_accepted_no_delete(self):
        self.args.action = 'accepted'
        self.args.name = id_generator()
        self.args.no_delete = True

        handle_feature_call(self.args, self.engine_mock)

        self.engine_mock.assert_has_calls([
            mock.call.accept_feature(
                name=self.args.name,
                delete_feature_branch=False,
            ),
        ])

    def test_list(self):
        self.args.action = "list"
        handle_feature_call(self.args, self.engine_mock)

        self.engine_mock.assert_has_calls([
            mock.call.list_features(),
        ])

    def test_undefined_action(self):
        self.args.action = id_generator()
        with self.assertRaises(RuntimeError):
            handle_feature_call(self.args, self.engine_mock)


class ReleaseCallTestCase(CoreTestCase):
    def test_start(self):
        self.args.action = "start"
        self.args.name = id_generator()

        with mock.patch('flowhub.core.do_hook') as patch:
            handle_release_call(self.args, self.engine_mock)

            patch.assert_has_calls([
                mock.call(self.args, self.engine_mock, "post-release-start", self.args.name),
            ])

            self.engine_mock.assert_has_calls([
                mock.call.start_release(name=self.args.name),
            ])

    def test_publish_with_name(self):
        self.args.action = "publish"
        self.args.no_cleanup = False
        self.args.name = id_generator()

        with mock.patch('flowhub.core.do_hook') as patch:
            handle_release_call(self.args, self.engine_mock)

            patch.assert_has_calls([
                mock.call(self.args, self.engine_mock, "pre-release-publish"),
                mock.call().__nonzero__(),  # the if check
                mock.call(self.args, self.engine_mock, "post-release-publish", mock.ANY),
            ])

            self.engine_mock.assert_has_calls([
                mock.call.publish_release(
                    name=self.args.name,
                    with_delete=not self.args.no_cleanup,
                    tag_info=TagInfo(
                        self.engine_mock.release.name.replace().strip(),
                        "",
                    )
                ),
            ])

    def test_publish_failed_hook(self):
        self.args.action = "publish"
        self.args.name = id_generator()

        with mock.patch('flowhub.core.do_hook') as patch:
            patch.return_value = False

            handle_release_call(self.args, self.engine_mock)

            patch.assert_has_calls([
                mock.call(self.args, self.engine_mock, "pre-release-publish"),
            ])

            self.assertEqual(self.engine_mock.call_count, 0)

    def test_contribute(self):
        self.args.action = "contribute"

        handle_release_call(self.args, self.engine_mock)

        self.engine_mock.assert_has_calls([
            mock.call.contribute_release(),
        ])


class HotfixCallTestCase(CoreTestCase):
    def test_start(self):
        self.args.action = "start"
        self.args.name = id_generator()
        self.args.issue_numbers = []

        with mock.patch('flowhub.core.do_hook') as patch:
            patch.return_value = True

            handle_hotfix_call(self.args, self.engine_mock)

            patch.assert_has_calls([
                mock.call(
                    self.args,
                    self.engine_mock,
                    "post-hotfix-start",
                    self.args.name,
                ),
            ])

            self.engine_mock.assert_has_calls([
                mock.call.start_hotfix(
                    name=self.args.name,
                    issues=self.args.issue_numbers
                )
            ])

    def test_publish_with_name(self):
        self.args.action = "publish"
        self.args.name = id_generator()
        self.args.issue_numbers = []
        with mock.patch('flowhub.core.do_hook') as patch:
            patch.return_value = True

            handle_hotfix_call(self.args, self.engine_mock)

            patch.assert_has_calls([
                mock.call(self.args, self.engine_mock, 'pre-hotfix-publish'),
                mock.call(self.args, self.engine_mock, 'post-hotfix-publish', mock.ANY),

            ])

            self.engine_mock.assert_has_calls([
                mock.call.publish_hotfix(
                    name=self.args.name,
                    tag_info=TagInfo(
                        self.engine_mock.hotfix.name.replace().strip(),
                        "",
                    )
                ),
            ])

    def test_contribute(self):
        self.args.action = "contribute"

        handle_hotfix_call(self.args, self.engine_mock)

        self.engine_mock.assert_has_calls([
            mock.call.contribute_hotfix(),
        ])


class CleanupCallTestCase(CoreTestCase):
    OPTIONS = ['t', 'u', 'r']

    def test_option_combinations(self):

        for l in range(1, len(self.OPTIONS)+1):

            for combo in itertools.combinations(self.OPTIONS, l):
                # reset the chosen options
                for o in self.OPTIONS:
                    setattr(self.args, o, False)
                self.args.all = False

                print combo
                for choice in combo:
                    setattr(self.args, choice, True)

                handle_cleanup_call(self.args, self.engine_mock)

                self.engine_mock.assert_has_calls([
                    mock.call.cleanup_branches(targets=''.join(combo))
                ])
                self.engine_patch.stop()
                self.engine_patch.start()

    def test_all(self):
        for o in self.OPTIONS:
            setattr(self.args, o, False)
        self.args.all = True

        handle_cleanup_call(self.args, self.engine_mock)

        self.engine_mock.assert_has_calls([
            mock.call.cleanup_branches(targets='tur')
        ])

    def test_no_targets(self):
        for o in self.OPTIONS:
            setattr(self.args, o, False)
        self.args.all = False

        handle_cleanup_call(self.args, self.engine_mock)

        self.assertEqual(self.engine_mock.call_count, 0)


class IssueCallTestCase(CoreTestCase):
    def test_start(self):
        self.args.action = "start"
        self.args.labels = False
        self.args.create_branch = bool(random.choice([0, 1]))

        handle_issue_call(self.args, self.engine_mock)

        self.engine_mock.assert_has_calls([
            mock.call.open_issue(
                title=self.args.title,
                labels=None,
                create_branch=self.args.create_branch,
            )
        ])

    def test_start_with_labels(self):
        self.args.action = "start"
        self.args.labels = ",".join((id_generator() for x in range(5)))
        self.args.create_branch = bool(random.choice([0, 1]))

        handle_issue_call(self.args, self.engine_mock)

        self.engine_mock.assert_has_calls([
            mock.call.open_issue(
                title=self.args.title,
                labels=self.args.labels.split(','),
                create_branch=self.args.create_branch,
            )
        ])



"""
Copyright (C) 2012 Haak Saxberg

This file is part of Flowhub, a command-line tool to enable various
Git-based workflows that interacts with GitHub.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import git
import mock
import unittest

from tests import (
    id_generator,
)

from flowhub.engine import Engine
from flowhub.managers import TagInfo


class NotARepoSetupTestCase(unittest.TestCase):
    def test_setup_abort(self):
        def raise_call_error(*args, **kwargs):
            raise git.exc.InvalidGitRepositoryError

        with mock.patch("git.Repo.__init__") as patch:
            patch.side_effect = raise_call_error

            with self.assertRaises(git.exc.InvalidGitRepositoryError):
                Engine(offline=True)


class EngineTestCase(unittest.TestCase):
    def setUp(self):
        self.gh_patch = mock.patch("flowhub.engine.Github")
        self.git_patch = mock.patch("git.Repo")
        self.configurator_patch = mock.patch("flowhub.engine.Configurator")

        self.gh_mock = self.gh_patch.start()
        self.git_mock = self.git_patch.start()
        self.configurator_mock = self.configurator_patch.start()

        self.feature_m_patch = mock.patch('flowhub.engine.FeatureManager')
        self.feature_m_mock = self.feature_m_patch.start()
        self.release_m_patch = mock.patch('flowhub.engine.ReleaseManager')
        self.release_m_mock = self.release_m_patch.start()
        self.hotfix_m_patch = mock.patch('flowhub.engine.HotfixManager')
        self.hotfix_m_mock = self.hotfix_m_patch.start()
        self.pull_m_patch = mock.patch('flowhub.engine.PullRequestManager')
        self.pull_m_mock = self.pull_m_patch.start()

    def tearDown(self):
        super(EngineTestCase, self).tearDown()
        self.pull_m_patch.stop()
        self.hotfix_m_patch.stop()
        self.release_m_patch.stop()
        self.feature_m_patch.stop()
        self.configurator_patch.stop()
        self.git_patch.stop()
        self.gh_patch.stop()


class OfflineTestCase(unittest.TestCase):
    def _produce_engine(self):
        engine = Engine(INIT=True, offline=True)

        self.repository_structure = {
            "name": id_generator(),
            "origin": id_generator(),
            "canon": id_generator(),
            "master": id_generator(),
            "develop": id_generator(),
            "feature": id_generator(),
            "release": id_generator(),
            "hotfix": id_generator(),
        }

        engine.setup_repository_structure(
            **self.repository_structure
        )

        # setup the configurator mock
        for key in ["name", "origin", "canon", "master", "develop"]:
            setattr(
                self.configurator_mock().flowhub.structure,
                key,
                self.repository_structure[key]
            )
        for key in ["feature", "release", "hotfix"]:
            setattr(
                self.configurator_mock().flowhub.prefix,
                key,
                self.repository_structure[key]
            )

        self.engine = Engine(offline=True)
        self.feature_m_mock.assert_has_calls([
            mock.call(
                debug=self.engine.DEBUG,
                prefix=self.repository_structure['feature'],
                origin=mock.ANY,
                canon=mock.ANY,
                master=mock.ANY,
                develop=mock.ANY,
                release=mock.ANY,
                hotfix=mock.ANY,
                repo=mock.ANY,
                gh=None,
                offline=True
            ),
        ])

        self.release_m_mock.assert_has_calls([
            mock.call(
                debug=self.engine.DEBUG,
                prefix=self.repository_structure['release'],
                origin=mock.ANY,
                canon=mock.ANY,
                master=mock.ANY,
                develop=mock.ANY,
                release=mock.ANY,
                hotfix=mock.ANY,
                repo=mock.ANY,
                gh=None,
                offline=True
            ),
        ])

        self.hotfix_m_mock.assert_has_calls([
            mock.call(
                debug=self.engine.DEBUG,
                prefix=self.repository_structure['hotfix'],
                origin=mock.ANY,
                canon=mock.ANY,
                master=mock.ANY,
                develop=mock.ANY,
                release=mock.ANY,
                hotfix=mock.ANY,
                repo=mock.ANY,
                gh=None,
                offline=True
            ),
        ])

        self.pull_m_mock.assert_has_calls([
            mock.call(
                debug=self.engine.DEBUG,
                prefix=self.repository_structure['name'],
                origin=mock.ANY,
                canon=mock.ANY,
                master=mock.ANY,
                develop=mock.ANY,
                release=mock.ANY,
                hotfix=mock.ANY,
                repo=mock.ANY,
                gh=None,
                offline=True
            ),
        ])

    def tearDown(self):
        # ensure no Github accesses have happened
        self.assertEqual(self.gh_mock.call_count, 0)
        super(OfflineTestCase, self).tearDown()


class OnlineTestCase(unittest.TestCase):
    def _produce_engine(self):
        engine = Engine(INIT=True)

        self.repository_structure = {
            "name": id_generator(),
            "origin": id_generator(),
            "canon": id_generator(),
            "master": id_generator(),
            "develop": id_generator(),
            "feature": id_generator(),
            "release": id_generator(),
            "hotfix": id_generator(),
        }

        engine.setup_repository_structure(
            **self.repository_structure
        )

        # setup the configurator mock
        for key in ["name", "origin", "canon", "master", "develop"]:
            setattr(
                self.configurator_mock().flowhub.structure,
                key,
                self.repository_structure[key]
            )
        for key in ["feature", "release", "hotfix"]:
            setattr(
                self.configurator_mock().flowhub.prefix,
                key,
                self.repository_structure[key]
            )

        self.engine = Engine()
        self.feature_m_mock.assert_has_calls([
            mock.call(
                debug=self.engine.DEBUG,
                prefix=self.repository_structure['feature'],
                origin=mock.ANY,
                canon=mock.ANY,
                master=mock.ANY,
                develop=mock.ANY,
                release=mock.ANY,
                hotfix=mock.ANY,
                repo=mock.ANY,
                gh=self.gh_mock(),
                offline=False
            ),
        ])

        self.release_m_mock.assert_has_calls([
            mock.call(
                debug=self.engine.DEBUG,
                prefix=self.repository_structure['release'],
                origin=mock.ANY,
                canon=mock.ANY,
                master=mock.ANY,
                develop=mock.ANY,
                release=mock.ANY,
                hotfix=mock.ANY,
                repo=mock.ANY,
                gh=self.gh_mock(),
                offline=False
            ),
        ])

        self.hotfix_m_mock.assert_has_calls([
            mock.call(
                debug=self.engine.DEBUG,
                prefix=self.repository_structure['hotfix'],
                origin=mock.ANY,
                canon=mock.ANY,
                master=mock.ANY,
                develop=mock.ANY,
                release=mock.ANY,
                hotfix=mock.ANY,
                repo=mock.ANY,
                gh=self.gh_mock(),
                offline=False
            ),
        ])

        self.pull_m_mock.assert_has_calls([
            mock.call(
                debug=self.engine.DEBUG,
                prefix=self.repository_structure['name'],
                origin=mock.ANY,
                canon=mock.ANY,
                master=mock.ANY,
                develop=mock.ANY,
                release=mock.ANY,
                hotfix=mock.ANY,
                repo=mock.ANY,
                gh=self.gh_mock(),
                offline=False
            ),
        ])


class OfflineFeatureTestCase(EngineTestCase, OfflineTestCase):
    def setUp(self):
        super(OfflineFeatureTestCase, self).setUp()
        self._produce_engine()

    def test_start_all_defaults(self):
        self.assertFalse(self.engine._create_feature())

    def test_start(self):
        name = id_generator()
        self.assertTrue(self.engine._create_feature(name))

        self.feature_m_mock.assert_has_calls([
            mock.call().start(name, True, mock.ANY),
        ])

    def test_start_with_tracking(self):
        name = id_generator()

        self.assertTrue(self.engine._create_feature(name, False))

        self.feature_m_mock.assert_has_calls([
            mock.call().start(name, False, mock.ANY),
        ])

    def test_work_all_defaults(self):
        self.assertFalse(self.engine.work_feature())

    def test_work_single_branch(self):
        name = id_generator()
        branch = mock.MagicMock()
        self.feature_m_mock().fuzzy_get.return_value = [branch]

        self.engine.work_feature(name)

        self.feature_m_mock.assert_has_calls([
            mock.call().fuzzy_get(name),
        ])
        branch.assert_has_calls([
            mock.call.checkout(),
        ])

    def test_work_multiple_branches(self):
        name = id_generator()
        branch = mock.MagicMock()
        self.feature_m_mock().fuzzy_get.return_value = [branch, branch]

        self.engine.work_feature(name)

        self.feature_m_mock.assert_has_calls([
            mock.call().fuzzy_get(name),
        ])
        self.assertEqual(branch.call_count, 0)

    def test_accept_all_defaults(self):
        self.assertFalse(self.engine._accept_feature())

    def test_accept(self):
        name = id_generator()
        return_branch = mock.MagicMock()
        self.git_mock().head.reference = return_branch

        self.assertTrue(self.engine._accept_feature(name))

        self.feature_m_mock.assert_has_calls([
            mock.call().accept(
                name,
                summary=mock.ANY,
                with_delete=True,
            ),
        ])

        return_branch.assert_has_calls([
            mock.call.checkout(),
        ])

    def test_accept_while_on_branch(self):
        name = id_generator()
        self.git_mock().head.reference.name = self.repository_structure['feature'] + name
        return_branch = self.engine.develop

        self.assertTrue(self.engine._accept_feature())

        self.feature_m_mock.assert_has_calls([
            mock.call().accept(name, summary=mock.ANY, with_delete=True),
        ])

        return_branch.assert_has_calls([
            mock.call.checkout(),
        ])

    def test_abandon_all_defaults(self):
        self.assertFalse(self.engine._abandon_feature())

    def test_abandon(self):
        name = id_generator()
        return_branch = mock.MagicMock()
        self.git_mock().head.reference = return_branch

        self.assertTrue(self.engine._abandon_feature(name))

        return_branch.assert_has_calls([
            mock.call.checkout(),
        ])

        self.feature_m_mock.assert_has_calls([
            mock.call().abandon(name, summary=mock.ANY),
        ])

    def test_abandon_while_on_branch(self):
        name = id_generator()
        self.git_mock().head.reference.name = self.repository_structure['feature'] + name
        return_branch = self.engine.develop

        self.assertTrue(self.engine._abandon_feature())

        return_branch.assert_has_calls([
            mock.call.checkout(),
        ])

        self.feature_m_mock.assert_has_calls([
            mock.call().abandon(name, summary=mock.ANY),
        ])

    def test_publish_all_defaults(self):
        self.assertFalse(self.engine._publish_feature())

    def test_publish(self):
        name = id_generator()

        self.assertFalse(self.engine._publish_feature(name))


class OnlineFeatureTestCase(EngineTestCase, OnlineTestCase):
    def setUp(self):
        super(OnlineFeatureTestCase, self).setUp()
        self._produce_engine()

    def test_publish(self):
        name = id_generator()
        b = mock.MagicMock()
        self.feature_m_mock().publish.return_value = b

        self.assertTrue(self.engine._publish_feature(name))

        self.feature_m_mock.assert_has_calls([
            mock.call().publish(name, mock.ANY),
        ])


class OfflineReleaseTestCase(EngineTestCase, OfflineTestCase):
    def setUp(self):
        super(OfflineReleaseTestCase, self).setUp()
        self._produce_engine()

    def test_start_all_defaults(self):
        self.assertFalse(self.engine._start_release())

    def test_start(self):
        name = id_generator()
        new_branch = mock.MagicMock()
        self.release_m_mock().start.return_value = new_branch
        self.assertTrue(self.engine._start_release(name))

        self.release_m_mock.assert_has_calls([
            mock.call().start(name, mock.ANY),
        ])

        new_branch.assert_has_calls([
            mock.call.checkout(),
        ])

    def test_start_with_existing_release(self):
        name = id_generator()

        release_mock = mock.MagicMock()
        release_mock.name.startswith.return_value = True
        self.git_mock().branches = [release_mock]

        self.assertFalse(self.engine._start_release(name))

        self.assertEqual(self.release_m_mock().call_count, 0)

    def test_publish_all_defaults(self):
        self.assertFalse(self.engine._publish_release())

    def test_publish_with_name(self):
        name = id_generator()

        self.assertTrue(self.engine._publish_release(name))

        self.release_m_mock.assert_has_calls([
            mock.call().publish(name, True, None, mock.ANY)
        ])

    def test_publish_on_release_branch(self):
        return_branch = self.engine.develop
        self.git_mock().head.reference.name = self.repository_structure['release'] + id_generator()

        self.assertTrue(self.engine._publish_release())

        return_branch.assert_has_calls([
            mock.call.checkout(),
        ])

    def test_publish_with_name_tag_info_and_no_delete(self):
        name = id_generator()
        tag_label = id_generator()
        tag_message = id_generator()

        self.assertTrue(
            self.engine._publish_release(
                name,
                with_delete=False,
                tag_info=TagInfo(tag_label, tag_message),
            ),
        )

        self.release_m_mock.assert_has_calls([
            mock.call().publish(name, False, TagInfo(tag_label, tag_message), mock.ANY)
        ])

    def test_contribute(self):
        self.assertFalse(self.engine.contribute_release())


class OnlineReleaseTestCase(EngineTestCase, OnlineTestCase):
    def setUp(self):
        super(OnlineReleaseTestCase, self).setUp()
        self._produce_engine()

    def test_contribute(self):
        release_mock = mock.MagicMock()
        release_mock.name.startswith.return_value = True
        release_mock.commit = True

        self.git_mock().branches = [release_mock]

        self.git_mock().head.reference.object.iter_parents.return_value = [True]

        self.assertTrue(self.engine._contribute_release())

        self.release_m_mock.assert_has_calls([
            mock.call().contribute(self.git_mock().head.reference, mock.ANY),
        ])


class OfflineHotfixTestCase(EngineTestCase, OfflineTestCase):
    def setUp(self):
        super(OfflineHotfixTestCase, self).setUp()
        self._produce_engine()

    def test_start_all_defaults(self):
        self.assertFalse(self.engine._start_hotfix())

    def test_start(self):
        name = id_generator()
        return_branch = self.hotfix_m_mock().start()

        self.assertTrue(self.engine._start_hotfix(name))

        self.hotfix_m_mock.assert_has_calls([
            mock.call().start(name, None, mock.ANY),
        ])

        return_branch.assert_has_calls([
            mock.call.checkout(),
        ])

    def test_start_with_issues(self):
        name = id_generator()
        issues = mock.MagicMock()
        return_branch = self.hotfix_m_mock().start()

        self.assertTrue(self.engine._start_hotfix(name, issues))

        self.hotfix_m_mock.assert_has_calls([
            mock.call().start(name, issues, mock.ANY),
        ])

        return_branch.assert_has_calls([
            mock.call.checkout(),
        ])

    def test_start_with_existing_hotfix(self):
        name = id_generator()

        hotfix_mock = mock.MagicMock()
        hotfix_mock.name.startswith.return_value = True
        self.git_mock().branches = [hotfix_mock]

        self.assertFalse(self.engine._start_hotfix(name))

        self.assertEqual(self.hotfix_m_mock().call_count, 0)

    def test_publish_all_defaults(self):
        self.assertFalse(self.engine._publish_hotfix())

    def test_publish_not_on_hotfix_branch(self):
        name = id_generator()

        return_branch = mock.MagicMock()
        self.git_mock().head.reference = return_branch

        self.assertTrue(self.engine._publish_hotfix(name))

        self.hotfix_m_mock.assert_has_calls([
            mock.call().publish(name, None, True, mock.ANY),
        ])

        return_branch.assert_has_calls([
            mock.call.checkout(),
        ])

    def test_publish_on_hotfix_branch(self):
        return_branch = self.engine.develop
        self.git_mock().head.reference.name = self.repository_structure['hotfix'] + id_generator()

        self.assertTrue(self.engine._publish_hotfix())

        return_branch.assert_has_calls([
            mock.call.checkout(),
        ])

    def test_contribute(self):
        self.assertFalse(self.engine._contribute_hotfix())


class OnlineHotfixTestCase(EngineTestCase, OnlineTestCase):
    def setUp(self):
        super(OnlineHotfixTestCase, self).setUp()
        self._produce_engine()

    def test_contribute(self):
        hotfix_mock = mock.MagicMock()
        hotfix_mock.name.startswith.return_value = True
        hotfix_mock.commit = True

        self.git_mock().branches = [hotfix_mock]

        self.git_mock().head.reference.object.iter_parents.return_value = [True]

        self.assertTrue(self.engine._contribute_hotfix())

        self.release_m_mock.assert_has_calls([
            mock.call().contribute(self.git_mock().head.reference, mock.ANY),
        ])

    def test_contribute_on_wrong_branch_by_existance(self):
        self.git_mock().branches = []
        self.assertFalse(self.engine._contribute_hotfix())

    def test_contribute_on_wrong_branch_by_commit(self):
        hotfix_mock = mock.MagicMock()
        hotfix_mock.name.startswith.return_value = True

        self.git_mock().branches = [hotfix_mock]

        # ensure that the "commit" isn't in the iter_parents
        hotfix_mock.commit = False
        self.git_mock().head.reference.object.iter_parents.return_value = [True]

        self.assertFalse(self.engine._contribute_hotfix())




#!/usr/bin/env python
"""
Copyright (C) 2012 Haak Saxberg

This file is part of Flowhub, a command-line tool to enable various
Git-based workflows that interacts with GitHub.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

from distutils.core import setup
from setuptools import find_packages
import os

setup(
    name="flowhub",
    version='0.5.1a',
    description="Git-flow adapted for GitHub",
    long_description=open("README.txt").read(),
    author="Haak Saxberg",
    author_email="haak.erling@gmail.com",
    url="http://github.com/haaksmash/flowhub",
    packages=find_packages(),
    scripts=[
        os.path.join('bin', 'flowhub')
    ],
    install_requires=[
        'GitPython == 0.3.2.RC1',
        'PyGithub > 1.14',
    ],
)



"""
Copyright (C) 2012 Haak Saxberg

This file is part of Flowhub, a command-line tool to enable various
Git-based workflows that interacts with GitHub.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import re

from github import GithubException

from flowhub.managers import Manager


def sanitize_refs(method):
    def wrapper(self, base, head, *args, **kwargs):
        if self.canon != self.origin:
            head = "{}:{}".format(self.gh.get_user().login, head)

        return method(self, base, head, *args, **kwargs)
    return wrapper


class PullRequestManager(Manager):

    def __init__(self, *args, **kwargs):
        super(PullRequestManager, self).__init__(*args, **kwargs)

        if self.offline:
            self.gh_repo = None

        else:
            if self.canon == self.origin:
                self.gh_repo = self.gh.get_user().get_repo(self._prefix)
            else:
                self.gh_repo = self.gh.get_user().get_repo(self._prefix).parent

    @sanitize_refs
    def create_from_branch_name(self, base, head, summary):

        if self.DEBUG > 1:
            print "setting up new pull-request"

        component_name = head.name.split('/')[-1]

        # check for issue-numbers at the front of the branch name
        # if there is one, attach the pull-request to that issue.
        match = re.match('^\d+', component_name)
        if match:
            issue_number = match.group()
            issue = self.gh_repo.get_issue(int(issue_number))
            pr = self.gh_repo.create_pull(
                issue=issue,
                base=base.name,
                head=head.name,
            )
            summary += [
                "New pull request created: {} into {}"
                "\n\turl: {}".format(
                    head,
                    base,
                    pr.issue_url,
                )
            ]
            return pr

        return False

    @sanitize_refs
    def create_pull(self, base, head, issue, summary):
        pr = self.gh_repo.create_pull(
            issue=issue,
            base=base.name,
            head=head.name,
        )

        summary += [
            "New pull request created: {} into {}"
            "\n\turl: {}".format(
                head,
                base,
                pr.issue_url,
            )
        ]

        return pr

    @sanitize_refs
    def add_to_pull(self, base, head, summary):
        if self.offline:
            return False

        prs = [
            x for x in self.gh_repo.get_pulls('open')
            if x.head.label == head
            or x.head.label == "{}:{}".format(self.gh.get_user().login, head)
        ]

        # If there's already a pull-request, don't bother hitting the gh api.
        if prs:
            pr = prs[0]
            summary += [
                "New commits added to existing pull-request"
                "\n\turl: {}".format(pr.issue_url)
            ]
            return pr

        else:
            return False

    def get_issue(self, issue_num):
        try:
            return self.gh_repo.get_issue(issue_num)
        except GithubException:
            return None

    def open_issue(self, title, body, labels, summary):
        gh_labels = [l for l in self.gh_repo.get_labels() if l.name in labels]

        issue = self.gh_repo.create_issue(
            title=title,
            body=body or "No description provided.",
            labels=gh_labels,
        )

        summary += [
            'Opened issue #{}: {}{}\n'
            '\turl: {}'.format(
                issue.number,
                title,
                '\n\t[{}]'.format(' '.join([l.name for l in gh_labels])) if gh_labels else '',
                issue.url,
            )
        ]

        return issue



"""
Copyright (C) 2012 Haak Saxberg

This file is part of Flowhub, a command-line tool to enable various
Git-based workflows that interacts with GitHub.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

from flowhub.managers import Manager


class FeatureManager(Manager):

    def start(self, name, with_tracking, summary):
        branch_name = "{}{}".format(
            self._prefix,
            name,
        )
        branch = self.repo.create_head(
            branch_name,
            commit=self.develop,  # Requires a develop branch.
        )
        summary += [
            "New branch {} created, from branch {}".format(
                branch_name,
                self.develop,
            )
        ]

        # can't create a tracking branch if we're offline
        if not self.offline and with_tracking:
            if self.DEBUG > 0:
                print "Adding a tracking branch to your GitHub repo"

            self.repo.git.push(
                self.origin,
                branch_name,
                set_upstream=True
            )
            summary += [
                "Created a remote tracking branch on {} for {}".format(
                    self.origin.name,
                    branch_name,
                ),
            ]

        return branch

    def get(self, name):
        try:
            return getattr(self.repo.branches, "{}{}".format(self._prefix, name))
        except AttributeError:
            return None

    def fuzzy_get(self, name):
        branch_name = "{}{}".format(
            self._prefix,
            name
        )

        try:
            branches = [getattr(self.repo.branches, branch_name)]
        except AttributeError:
            branches = [b for b in self.repo.branches if b.startswith(branch_name)]

        return branches

    def accept(self, name, summary, with_delete):
        self.canon.fetch()
        summary += [
            "Latest objects fetched from {}".format(self.canon),
        ]
        self.develop.checkout()
        self.repo.git.merge(
            "{}/{}".format(self.canon, self.develop),
        )
        summary += [
            "Updated {}".format(self.develop),
        ]

        branch_name = "{}{}".format(
            self._prefix,
            name,
        )

        if with_delete:
            self.repo.delete_head(
                branch_name,
            )
            summary += [
                "Deleted {} from local repository".format(branch_name),
            ]

            if not self.offline:
                self.origin.push(
                    branch_name,
                    delete=True,
                )
                summary += [
                    "Deleted {} from {}".format(branch_name, self.origin),
                ]

    def abandon(self, name, summary):
        branch_name = "{}{}".format(
            self._prefix,
            name,
        )

        self.repo.delete_head(
            branch_name,
            force=True,
        )
        summary += [
            "Deleted branch {} locally".format(
                branch_name,
            ),
        ]

        if not self.offline:
            self.repo.git.push(
                self.origin,
                branch_name,
                delete=True,
                force=True,
            )
            summary[-1] += "and from remote {}".format(
                self.origin,
            )

    def publish(self, name, summary):
        branch_name = "{}{}".format(
            self._prefix,
            name,
        )
        self.repo.git.push(
            self.origin,
            branch_name,
            set_upstream=True,
        )
        summary += [
            "Updated {}/{}".format(self.origin, branch_name)
        ]

        return self.get(name)
